﻿using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Content;

namespace Engine
{
    /// <summary>
    /// First person camera component for the demos, rotated by mouse.
    /// </summary>
    public class FPSCamera : GameComponent
    {
        private Matrix view;
        private Matrix projection;

        KeyboardState keyState;
        KeyboardState prevKeyState;

        MouseState mouseState;
        MouseState prevMouseState;

        public Game game;

        Boolean isMoving = false;

        private Vector3 position = new Vector3(0, 0, 0);
        private Vector2 angles = Vector2.Zero;

        private float fieldOfView = Microsoft.Xna.Framework.MathHelper.PiOver4;
        private float aspectRatio;
        private float nearPlaneDistance = 0.1f;
        private float farPlaneDistance = 100000.0f;


        public FPSCamera(Game game)
            : base(game)
        {
            this.game = game;
            aspectRatio = (float)game.GraphicsDevice.Viewport.Width / (float)game.GraphicsDevice.Viewport.Height;
            UpdateProjection();
            Frustum = new BoundingFrustum(ViewProjection);
        }

        public BoundingFrustum Frustum { get; set; }
        /// <summary>
        /// Gets camera view matrix.
        /// </summary>
        public Matrix View { get { return view; } }
        /// <summary>
        /// Gets or sets camera projection matrix.
        /// </summary>
        public Matrix Projection { get { return projection; } set { projection = value; } }
        /// <summary>
        /// Gets camera view matrix multiplied by projection matrix.
        /// </summary>
        public Matrix ViewProjection { get { return view * projection; } }

        /// <summary>
        /// Gets or sets camera position.
        /// </summary>
        public Vector3 Position { get { return position; } set { position = value; } }

        /// <summary>
        /// Gets or sets camera field of view.
        /// </summary>
        public float FieldOfView { get { return fieldOfView; } set { fieldOfView = value; UpdateProjection(); } }
        /// <summary>
        /// Gets or sets camera aspect ratio.
        /// </summary>
        public float AspectRatio { get { return aspectRatio; } set { aspectRatio = value; UpdateProjection(); } }
        /// <summary>
        /// Gets or sets camera near plane distance.
        /// </summary>
        public float NearPlaneDistance { get { return nearPlaneDistance; } set { nearPlaneDistance = value; UpdateProjection(); } }
        /// <summary>
        /// Gets or sets camera far plane distance.
        /// </summary>
        public float FarPlaneDistance { get { return farPlaneDistance; } set { farPlaneDistance = value; UpdateProjection(); } }

        public Vector2 Angles { get { return angles; } set { angles = value; } }

        /// <summary>
        /// Gets or sets camera's target.
        /// </summary>
        public Vector3 Target
        {
            get
            {
                Matrix cameraRotation = Matrix.CreateRotationX(angles.X) * Matrix.CreateRotationY(angles.Y);
                return position + Vector3.Transform(Vector3.Forward, cameraRotation);
            }
            set
            {
                Vector3 forward = Vector3.Normalize(position - value);
                Vector3 right = Vector3.Normalize(Vector3.Cross(forward, Vector3.Up));
                Vector3 up = Vector3.Normalize(Vector3.Cross(right, forward));

                Matrix test = Matrix.Identity;
                test.Forward = forward;
                test.Right = right;
                test.Up = up;
                angles.X = -(float)Math.Asin(test.M32);
                angles.Y = -(float)Math.Asin(test.M13);
            }
        }

        /// <summary>
        /// Updates camera with input and updates view matrix.
        /// </summary>
        /// <param name="gameTime"></param>
        public override void Update(GameTime gameTime)
        {
            double elapsedTime = (double)gameTime.ElapsedGameTime.Ticks / (double)TimeSpan.TicksPerSecond;
            ProcessMouseAndKeyboardInput((float)elapsedTime * 1.0f);
            UpdateView();

            base.Update(gameTime);
        }

        Vector2 clickLoc = new Vector2();
        private void ProcessMouseAndKeyboardInput(float amountOfMovement)
        {
            prevKeyState = keyState;
            keyState = Keyboard.GetState();

            prevMouseState = mouseState;
            mouseState = Mouse.GetState();


            Vector3 moveVector = new Vector3();

            if (mouseState.RightButton == ButtonState.Pressed && prevMouseState.RightButton != ButtonState.Pressed)
            {
                clickLoc = new Vector2(mouseState.X, mouseState.Y);
                isMoving = true;
            }


            if (isMoving)
            {
                Mouse.SetPosition((int)clickLoc.X, (int)clickLoc.Y);//game.GraphicsDevice.Viewport.Width / 2, game.GraphicsDevice.Viewport.Height / 2);

                if (KeyDown(Keys.D))
                    moveVector.X += amountOfMovement;
                if (KeyDown(Keys.A))
                    moveVector.X -= amountOfMovement;
                if (KeyDown(Keys.S))
                    moveVector.Z += amountOfMovement;
                if (KeyDown(Keys.W))
                    moveVector.Z -= amountOfMovement;

                Matrix cameraRotation = Matrix.CreateRotationX(angles.X) * Matrix.CreateRotationY(angles.Y);
                position += Vector3.Transform(moveVector, cameraRotation);

                angles.Y += amountOfMovement / 2 * mouseDistFromCenter().X / 3;
                angles.X += amountOfMovement / 2 * mouseDistFromCenter().Y / 3;

                if (angles.X > 1.55f) angles.X = 1.55f;
                if (angles.X < -1.55f) angles.X = -1.55f;
                if (angles.Y > Math.PI) angles.Y -= 2 * (float)Math.PI;
                if (angles.Y < -Math.PI) angles.Y += 2 * (float)Math.PI;
            }

            if (mouseState.RightButton == ButtonState.Released && prevMouseState.RightButton == ButtonState.Pressed)
            {
                isMoving = false;
            }
        }

        private Boolean KeyDown(Keys key)
        {
            return keyState.IsKeyDown(key);
        }

        private Vector2 mouseDistFromCenter()
        {
            return (clickLoc - new Vector2(mouseState.X, mouseState.Y));
        }

        private void UpdateProjection()
        {
            projection = Matrix.CreatePerspectiveFieldOfView(fieldOfView, aspectRatio, nearPlaneDistance, farPlaneDistance);
        }

        private void UpdateView()
        {
            Matrix cameraRotation = Matrix.CreateRotationX(angles.X) * Matrix.CreateRotationY(angles.Y);
            Vector3 targetPos = position + Vector3.Transform(Vector3.Forward, cameraRotation);

            Vector3 upVector = Vector3.Transform(Vector3.Up, cameraRotation);

            view = Matrix.CreateLookAt(position, targetPos, upVector);

            Frustum.Matrix = ViewProjection;
        }
    }
}
