/*****************************************************************
 * Jackson Fields
 * Input.cs
 *****************************************************************/
using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using System.Windows.Forms;
using System.Drawing;
using Point = Microsoft.Xna.Framework.Point;
using Keys = Microsoft.Xna.Framework.Input.Keys;
using ButtonState = Microsoft.Xna.Framework.Input.ButtonState;

namespace Engine
{
    /// <summary>
    /// Handles input devices
    /// Maintains the current and previous state of each device
    /// 
    /// Derives from GameComponent - add this component to the
    /// game's component list
    /// </summary>
    public class Input : GameComponent
    {
        Form gameForm;
        Game game;

        KeyboardState keyState, prevKeyState;
        MouseState mouseState, prevMouseState;
        public Vector2 mouseLocation;
        public Point mousePoint;
        public Vector2 prevMouseLocation;
        public Point prevMousePoint;

        GamePadState padState1, prevPadState1;
        GamePadState padState2, prevPadState2;
        GamePadState padState3, prevPadState3;
        GamePadState padState4, prevPadState4;


        public Input(Game game, Form gameForm = null)
            : base(game)
        {
            this.gameForm = gameForm;
            this.game = game;
        }

        public Boolean inBounds()
        {
            return mouseLocation.X > 0 && mouseLocation.X < game.GraphicsDevice.Viewport.Width &&
                mouseLocation.Y > 0 && mouseLocation.Y < game.GraphicsDevice.Viewport.Height;
        }

        private Vector2 GetMouseLocInWindow(Vector2 pos)
        {
            System.Drawing.Point p = gameForm.PointToClient(new System.Drawing.Point((int)pos.X, (int)pos.Y));
            return new Vector2(p.X, p.Y);
        }

        private Point MousePoint(Vector2 pos)
        {
            if (gameForm != null)
            {
                Vector2 loc = GetMouseLocInWindow(pos);
                return new Point((int)loc.X, (int)loc.Y);
            }

            return new Point((int)pos.X, (int)pos.Y);
        }

        private Vector2 MouseLocation(Vector2 pos)
        {
            if (gameForm != null)
            {
                Vector2 loc = GetMouseLocInWindow(pos);
                return loc;
            }

            return pos;
        }

        public override void Update(GameTime gameTime)
        {
            prevKeyState = keyState;
            keyState = Keyboard.GetState();

            prevMouseState = mouseState;
            mouseState = Mouse.GetState();

            mouseLocation = MouseLocation(new Vector2(mouseState.X, mouseState.Y));
            mousePoint = MousePoint(new Vector2(mouseState.X, mouseState.Y));

            prevMouseLocation = MouseLocation(new Vector2(prevMouseState.X, prevMouseState.Y));
            prevMousePoint = MousePoint(new Vector2(prevMouseState.X, prevMouseState.Y));

            prevPadState1 = padState1;
            padState1 = GamePad.GetState(PlayerIndex.One);
            prevPadState2 = padState2;
            padState2 = GamePad.GetState(PlayerIndex.Two);
            prevPadState3 = padState3;
            padState3 = GamePad.GetState(PlayerIndex.Three);
            prevPadState4 = padState4;
            padState4 = GamePad.GetState(PlayerIndex.Four);
        }

        public Boolean ConfirmationInput()
        {
            if (IsControllerConnected())
            {
                PlayerIndex activePlayerIndex = PlayerIndex.One;
                Boolean gamePadEscapeInput = false;
                for (int i = 0; i < 4; i++)
                {
                    if (i == 0)
                    {
                        activePlayerIndex = PlayerIndex.One;
                    }
                    else if (i == 1)
                    {
                        activePlayerIndex = PlayerIndex.Two;
                    }
                    else if (i == 2)
                    {
                        activePlayerIndex = PlayerIndex.Three;
                    }
                    else if (i == 3)
                    {
                        activePlayerIndex = PlayerIndex.Four;
                    }

                    gamePadEscapeInput = Click(activePlayerIndex, Buttons.A) || Click(activePlayerIndex, Buttons.Start);

                    if (gamePadEscapeInput)
                    {
                        return true;
                    }
                }
            }

            return (Click(Keys.Enter));
        }

        public Boolean EscapeInput()
        {
            // (IsControllerConnected())
            //{
                PlayerIndex activePlayerIndex = PlayerIndex.One;
                Boolean gamePadEscapeInput = false;
                for (int i = 0; i < 4; i++)
                {
                    if (i == 0)
                    {
                        activePlayerIndex = PlayerIndex.One;
                    }
                    else if (i == 1)
                    {
                        activePlayerIndex = PlayerIndex.Two;
                    }
                    else if (i == 2)
                    {
                        activePlayerIndex = PlayerIndex.Three;
                    }
                    else if (i == 3)
                    {
                        activePlayerIndex = PlayerIndex.Four;
                    }

                    gamePadEscapeInput = Click(activePlayerIndex, Buttons.Back);

                    if (gamePadEscapeInput)
                    {
                        return true;
                    }
                }
            //}

            return (Click(Keys.Escape));
        }

        #region Mouse
        public Boolean LeftClick()
        {
            return mouseState.LeftButton == ButtonState.Pressed &&
                prevMouseState.LeftButton != ButtonState.Pressed;
        }

        public Boolean RightClick()
        {
            return mouseState.RightButton == ButtonState.Pressed &&
                prevMouseState.RightButton != ButtonState.Pressed;
        }

        public Boolean MiddleClick()
        {
            return mouseState.MiddleButton == ButtonState.Pressed &&
                prevMouseState.MiddleButton != ButtonState.Pressed;
        }

        public Boolean LeftRelease()
        {
            return mouseState.LeftButton == ButtonState.Released &&
                prevMouseState.LeftButton != ButtonState.Released;
        }

        public Boolean RightRelease()
        {
            return mouseState.RightButton == ButtonState.Released &&
                prevMouseState.RightButton != ButtonState.Released;
        }

        public Boolean MiddleRelease()
        {
            return mouseState.MiddleButton == ButtonState.Released &&
                prevMouseState.MiddleButton != ButtonState.Released;
        }

        public Boolean LeftPressed()
        {
            return mouseState.LeftButton == ButtonState.Pressed;
        }

        public Boolean RightPressed()
        {
            return mouseState.RightButton == ButtonState.Pressed;
        }

        public Boolean MiddlePressed()
        {
            return mouseState.MiddleButton == ButtonState.Pressed;
        }
        #endregion

        #region Keyboard
        public Boolean Click(Keys key)
        {
            return keyState.IsKeyDown(key) &&
                !prevKeyState.IsKeyDown(key);
        }

        public Boolean IsKeyDown(Keys key)
        {
            return keyState.IsKeyDown(key);
        }

        public Boolean IsKeyUp(Keys key)
        {
            return keyState.IsKeyUp(key);
        }
        #endregion

        #region Gamepad
        public Boolean Click(PlayerIndex index, Buttons button)
        {
            switch (index)
            {
                case PlayerIndex.One:
                    return padState1.IsButtonDown(button) && 
                        !prevPadState1.IsButtonDown(button);
                case PlayerIndex.Two:
                    return padState2.IsButtonDown(button) &&
                        !prevPadState2.IsButtonDown(button);
                case PlayerIndex.Three:
                    return padState3.IsButtonDown(button) &&
                        !prevPadState3.IsButtonDown(button);
                case PlayerIndex.Four:
                    return padState4.IsButtonDown(button) &&
                        !prevPadState4.IsButtonDown(button);
            }

            return false;
        }

        public Boolean Click(Buttons button)
        {
            for (int i = 0; i < 4; i++)
            {
                switch (i)
                {
                    case 0:
                        return padState1.IsButtonDown(button) &&
                            !prevPadState1.IsButtonDown(button);
                    case 1:
                        return padState2.IsButtonDown(button) &&
                            !prevPadState2.IsButtonDown(button);
                    case 2:
                        return padState3.IsButtonDown(button) &&
                            !prevPadState3.IsButtonDown(button);
                    case 3:
                        return padState4.IsButtonDown(button) &&
                            !prevPadState4.IsButtonDown(button);
                }
            }

            return false;
        }

        public Boolean IsConnected(PlayerIndex index)
        {
            return GamePad.GetState(index).IsConnected;
        }

        public Boolean IsControllerConnected()
        {
            return padState1.IsConnected ||
                padState2.IsConnected ||
                padState3.IsConnected ||
                padState4.IsConnected;
        }

        public Boolean ControllerActionInput(out PlayerIndex? active)
        {
            PlayerIndex? activeIndex = null;
            for (int i = 0; i < 4; i++)
            {
                if (i == 0)
                {
                    activeIndex = PlayerIndex.One;
                    if (Click((PlayerIndex)activeIndex, Buttons.A) || Click((PlayerIndex)activeIndex, Buttons.Start))
                    {
                        active = activeIndex;
                        return true;
                    }
                }
                else if (i == 1)
                {
                    activeIndex = PlayerIndex.Two;
                    if (Click((PlayerIndex)activeIndex, Buttons.A) || Click((PlayerIndex)activeIndex, Buttons.Start))
                    {
                        active = activeIndex;
                        return true;
                    }
                }
                if (i == 3)
                {
                    activeIndex = PlayerIndex.Three;
                    if (Click((PlayerIndex)activeIndex, Buttons.A) || Click((PlayerIndex)activeIndex, Buttons.Start))
                    {
                        active = activeIndex;
                        return true;
                    }
                }
                if (i == 4)
                {
                    activeIndex = PlayerIndex.Four;
                    if (Click((PlayerIndex)activeIndex, Buttons.A) || Click((PlayerIndex)activeIndex, Buttons.Start))
                    {
                        active = activeIndex;
                        return true;
                    }
                }
            }

            active = activeIndex;
            return false;
        }
        #endregion
    }
}
