﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace Engine
{
    public static class Rays
    {
        public static Ray CalculateCenterRay(GraphicsDevice graphics, FPSCamera camera)
        {
            Vector2 center = new Vector2(graphics.Viewport.Width / 2, graphics.Viewport.Height / 2);
            Vector3 nearSource = new Vector3(center, 0f);
            Vector3 farSource = new Vector3(center, 1f);

            Vector3 nearPoint = graphics.Viewport.Unproject(nearSource,
                camera.Projection, camera.View, Matrix.Identity);

            Vector3 farPoint = graphics.Viewport.Unproject(farSource,
                camera.Projection, camera.View, Matrix.Identity);

            Vector3 direction = farPoint - nearPoint;
            direction.Normalize();

            return new Ray(nearPoint, direction);
        }

        public static Ray CalculateCursorRay(Vector2 mousePos, GraphicsDevice graphics, FPSCamera camera)
        {
            Vector3 nearSource = new Vector3(mousePos, 0f);
            Vector3 farSource = new Vector3(mousePos, 1f);

            Vector3 nearPoint = graphics.Viewport.Unproject(nearSource,
                camera.Projection, camera.View, Matrix.Identity);

            Vector3 farPoint = graphics.Viewport.Unproject(farSource,
                camera.Projection, camera.View, Matrix.Identity);

            Vector3 direction = farPoint - nearPoint;
            direction.Normalize();

            return new Ray(nearPoint, direction);
        }

        public static Vector3 CalculateWorldPosFromVector(Vector2 mousePos, GraphicsDevice graphics, FPSCamera camera)
        {
            Vector2 center = mousePos;

            Viewport vp = graphics.Viewport;
            //  Note the order of the parameters! Projection first.
            Vector3 pos1 = vp.Unproject(new Vector3(center.X, center.Y, 0), camera.Projection, camera.View, Matrix.Identity);
            Vector3 pos2 = vp.Unproject(new Vector3(center.X, center.Y, 1), camera.Projection, camera.View, Matrix.Identity);
            Vector3 dir = Vector3.Normalize(pos2 - pos1);

            if (dir.Y != 0)
            {
                Vector3 x = pos1 - dir * (pos1.Y / dir.Y);

                return x;
            }

            else return Vector3.Zero;
        }

        public static Vector3 CalculateWorldFromVectorWithPlane(Vector2 mousePos, GraphicsDevice graphics, FPSCamera camera, Plane plane)
        {
            Vector2 center = mousePos;
            Viewport vp = graphics.Viewport;

            Vector3 v1 = vp.Unproject(new Vector3(center.X, center.Y, 0), camera.Projection, camera.View, Matrix.Identity);
            Vector3 v2 = vp.Unproject(new Vector3(center.X, center.Y, 1), camera.Projection, camera.View, Matrix.Identity);
            Vector3 dir = Vector3.Normalize(v1 - v2);

            float angle = Vector3.Dot(plane.Normal, dir);
            // Part 1: Find if the ray is parallel to the plane by checking 
            //  if the angle between them is zero. 
            if (angle != 0)
            {
                float v1d = Vector3.Dot(plane.Normal, v1);
                // Part 2: Extend the starting point to the distance it would 
                //  require to intersect the plane. 
                return v1 + dir * ((plane.D - v1d) / angle);
            }

            return Vector3.Zero; 
        }

        public static Vector3 CalculateWorldFromVectorCameraHeight(Vector2 mousePos, GraphicsDevice graphics, FPSCamera camera)
        {
            Plane plane = new Plane(Vector3.Up, camera.Position.Y - 10f);

            Vector2 center = mousePos;
            Viewport vp = graphics.Viewport;

            Vector3 v1 = vp.Unproject(new Vector3(center.X, center.Y, 0), camera.Projection, camera.View, Matrix.Identity);
            Vector3 v2 = vp.Unproject(new Vector3(center.X, center.Y, 1), camera.Projection, camera.View, Matrix.Identity);
            Vector3 dir = Vector3.Normalize(v1 - v2);

            float angle = Vector3.Dot(plane.Normal, dir);
            // Part 1: Find if the ray is parallel to the plane by checking 
            //  if the angle between them is zero. 
            if (angle != 0)
            {
                float v1d = Vector3.Dot(plane.Normal, v1);
                // Part 2: Extend the starting point to the distance it would 
                //  require to intersect the plane. 
                return v1 + dir * ((plane.D - v1d) / angle);
            }

            return Vector3.Zero;
        }

        public static Vector3 WorldPositionOnTickWithPlane(Vector2 mousePos, Vector2 prevMousePos, GraphicsDevice graphics, FPSCamera camera, Plane plane)
        {
            Vector3 x;
            Vector3 px;

            Vector2 center = mousePos;
            Vector2 pcenter = prevMousePos;

            Viewport vp = graphics.Viewport;

            Vector3 ppos1 = vp.Unproject(new Vector3(pcenter.X, pcenter.Y, 0), camera.Projection, camera.View, Matrix.Identity);
            Vector3 ppos2 = vp.Unproject(new Vector3(pcenter.X, pcenter.Y, 1), camera.Projection, camera.View, Matrix.Identity);
            Vector3 pdir = Vector3.Normalize(ppos2 - ppos1);

            float pangle = Vector3.Dot(plane.Normal, pdir);

            if (pangle != 0)
            {
                float pv1d = Vector3.Dot(plane.Normal, ppos1);
                // Part 2: Extend the starting point to the distance it would 
                //  require to intersect the plane. 
                px = ppos1 + pdir * ((plane.D - pv1d) / pangle);
            }
            else
            {
                px = Vector3.Zero;
            }


            Vector3 pos1 = vp.Unproject(new Vector3(center.X, center.Y, 0), camera.Projection, camera.View, Matrix.Identity);
            Vector3 pos2 = vp.Unproject(new Vector3(center.X, center.Y, 1), camera.Projection, camera.View, Matrix.Identity);
            Vector3 dir = Vector3.Normalize(pos2 - pos1);

            float angle = Vector3.Dot(plane.Normal, dir);

            if (angle != 0)
            {
                float v1d = Vector3.Dot(plane.Normal, pos1);
                // Part 2: Extend the starting point to the distance it would 
                //  require to intersect the plane. 
                x = pos1 + dir * ((plane.D - v1d) / angle);
            }
            else
            {
                x = Vector3.Zero;
            }

            return px - x;
        }

        public static Vector3 WorldPositionOnTick(Vector2 mousePos, Vector2 prevMousePos, GraphicsDevice graphics, FPSCamera camera)
        {
            Vector3 x;
            Vector3 px;

            Vector2 center = mousePos;
            Vector2 pcenter = prevMousePos;

            Viewport vp = graphics.Viewport;
            
            Vector3 ppos1 = vp.Unproject(new Vector3(pcenter.X, pcenter.Y, 0), camera.Projection, camera.View, Matrix.Identity);
            Vector3 ppos2 = vp.Unproject(new Vector3(pcenter.X, pcenter.Y, 1), camera.Projection, camera.View, Matrix.Identity);
            Vector3 pdir = Vector3.Normalize(ppos2 - ppos1);

            if (pdir.Y != 0)
            {
                px = ppos1 - pdir * (ppos1.Y / pdir.Y);
            }
            else
            {
                px = Vector3.Zero;
            }


            Vector3 pos1 = vp.Unproject(new Vector3(center.X, center.Y, 0), camera.Projection, camera.View, Matrix.Identity);
            Vector3 pos2 = vp.Unproject(new Vector3(center.X, center.Y, 1), camera.Projection, camera.View, Matrix.Identity);
            Vector3 dir = Vector3.Normalize(pos2 - pos1);

            if (dir.Y != 0)
            {
                x = pos1 - dir * (pos1.Y / dir.Y);
            }
            else 
            {
                x = Vector3.Zero;
            }

            return px - x;
        }

        public static ModelMesh RayIntersectsMesh(FPSCamera camera, Ray ray, Model model, Matrix World)
        {
            foreach (ModelMesh mesh in model.Meshes)
            {
                BoundingSphere sphere = mesh.BoundingSphere;
                sphere = sphere.Transform(World);

                if (sphere.Intersects(ray) != null)
                {
                    List<Vector3> vertices = new List<Vector3>();
                    List<ModelMethods.TriangleVertexIndices> indices = new List<ModelMethods.TriangleVertexIndices>();

                    ModelMethods.ExtractModelMeshData(mesh, ref World, vertices, indices);

                    //make sure the vertices are translated by the object's position
                    for (int i = 0; i < indices.Count; i++)
                    {
                        float? intersection;
                        RayIntersectsTriangle(ref ray,
                                     vertices[indices[i].I0],
                                     vertices[indices[i].I1],
                                     vertices[indices[i].I2],
                                     out intersection);

                        if (intersection != null)
                        {
                            return mesh;
                        }
                    }
                }
            }
            return null;
        }

        //TODO: pick top-most object (not bottom)
        static void RayIntersectsTriangle(ref Ray ray, Vector3 vertex1,
                                          Vector3 vertex2,
                                          Vector3 vertex3, out float? result)
        {
            // Compute vectors along two edges of the triangle.
            Vector3 edge1, edge2;

            Vector3.Subtract(ref vertex2, ref vertex1, out edge1);
            Vector3.Subtract(ref vertex3, ref vertex1, out edge2);

            // Compute the determinant.
            Vector3 directionCrossEdge2;
            Vector3.Cross(ref ray.Direction, ref edge2, out directionCrossEdge2);

            float determinant;
            Vector3.Dot(ref edge1, ref directionCrossEdge2, out determinant);

            // If the ray is parallel to the triangle plane, there is no collision.
            if (determinant > -float.Epsilon && determinant < float.Epsilon)
            {
                result = null;
                return;
            }

            float inverseDeterminant = 1.0f / determinant;

            // Calculate the U parameter of the intersection point.
            Vector3 distanceVector;
            Vector3.Subtract(ref ray.Position, ref vertex1, out distanceVector);

            float triangleU;
            Vector3.Dot(ref distanceVector, ref directionCrossEdge2, out triangleU);
            triangleU *= inverseDeterminant;

            // Make sure it is inside the triangle.
            if (triangleU < 0 || triangleU > 1)
            {
                result = null;
                return;
            }

            // Calculate the V parameter of the intersection point.
            Vector3 distanceCrossEdge1;
            Vector3.Cross(ref distanceVector, ref edge1, out distanceCrossEdge1);

            float triangleV;
            Vector3.Dot(ref ray.Direction, ref distanceCrossEdge1, out triangleV);
            triangleV *= inverseDeterminant;

            // Make sure it is inside the triangle.
            if (triangleV < 0 || triangleU + triangleV > 1)
            {
                result = null;
                return;
            }

            // Compute the distance along the ray to the triangle.
            float rayDistance;
            Vector3.Dot(ref edge2, ref distanceCrossEdge1, out rayDistance);
            rayDistance *= inverseDeterminant;

            // Is the triangle behind the ray origin?
            if (rayDistance < 0)
            {
                result = null;
                return;
            }

            result = rayDistance;
        }
    }
}
