﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Engine;

namespace kinectApp
{
    public class Tree
    {
        PointCloud pointCloud;
        Game1 game;

        BranchCreator branchCreator;

        TreeRules treeRules;

        #region Leaves
        //leaves
        private List<LeafTexture> leafQuads = new List<LeafTexture>();
        VertexDeclaration vertexDeclaration;
        BasicEffect quadEffect;
        #endregion

        #region branches
        public List<VertexPositionNormal> branchVertices = new List<VertexPositionNormal>();
        public List<int> branchIndices = new List<int>();

        public VertexBuffer branchVertexBuffer;
        public IndexBuffer branchIndexBuffer;
        public BasicEffect branchBasicEffect;
        #endregion

        public Tree(Game game, BranchCreator branchCreator, PointCloud pointCloud, Boolean initialize = true)
        {
            this.game = (Game1)game;
            treeRules = this.game.TreeRules();
            this.branchCreator = branchCreator;
            this.pointCloud = pointCloud;

            quadEffect = new BasicEffect(game.GraphicsDevice);

            vertexDeclaration = new VertexDeclaration(new VertexElement[]
                {
                    new VertexElement(0, VertexElementFormat.Vector3, VertexElementUsage.Position, 0),
                    new VertexElement(12, VertexElementFormat.Vector3, VertexElementUsage.Normal, 0),
                    new VertexElement(24, VertexElementFormat.Vector2, VertexElementUsage.TextureCoordinate, 0)
                }
            );

            if (initialize)
            {
                Initialize();
            }
        }

        public void Initialize()
        {
            starts = new List<Vector3>();
            FindTrunkNodes();
            FindLeafLobes();
            ConnectTrunksToLeafLobes();
            FillLeafLobes();

            foreach (TreeBranch branch in branchCreator.treeHierarchy.GetTerminatingBranches())
            {
                //branch.isTrunk = false;
                //branch.SetBranchType(BranchType.Branch);

                branch.AddLeaves();
            }

            CreateLeaves();
        }

        /// <summary>
        /// Fill lobes and create leaves
        /// Invoke after preliminary branches and Lobes are set
        /// </summary>
        public void ResetLobes(BranchCreator branchCreator)
        {
            Console.WriteLine("Resetting Tree");
            this.branchCreator = branchCreator;
            ConnectTrunksToLeafLobes();
            FillLeafLobes();
            CreateLeaves();
        }

        //find apparent trunk lines from point cloud
        public void FindTrunkNodes()
        {
            Console.WriteLine("Identifying Trunk Nodes");
            //lineManager.AddLine(new Line(Vector3.Zero, Vector3.Up * 3));
        }

        //find areas in point cloud where leaves should be
        public void FindLeafLobes()
        {
            Console.WriteLine("Finding Leaf Lobes");
        }

        //connect the existing trunk(s) to leaf lobes
        public void ConnectTrunksToLeafLobes()
        {
            Console.WriteLine("Connecting trunks to Leaf Lobes");

            List<TreeBranch> terminatingBranches = branchCreator.treeHierarchy.GetTerminatingBranches();
            //List<TreeBranch> allBranches = branchCreator.treeHierarchy.GetAllBranches();
            foreach (TreeBranch branch in terminatingBranches)
            {
                //branch.isTrunk = true;
                //branch.SetBranchType(BranchType.Trunk);
                LeafLobe closestLobe = null;
                float closestDistance = float.MaxValue;
                Boolean branchContainedInLobe = false;
                foreach (LeafLobe lobe in game.leafCreator.spheres)
                {
                    float distance = Vector3.Distance(branch.end, lobe.sphere.Center);
                    if (distance < closestDistance)
                    {
                        closestDistance = distance;
                        closestLobe = lobe;
                    }

                    if (lobe.sphere.Contains(branch.end) != ContainmentType.Disjoint)
                    {
                        //branchContainedInLobe = true;
                        branch.isInsideLobe = true;
                        branch.leafLobe = lobe;
                        lobe.branchesInLobe.Add(branch);
                        break;
                    }
                }

                //if (!branchContainedInLobe)
                //{
                    if (closestLobe != null)
                    {
                        branch.leafLobe = closestLobe;
                        closestLobe.branchesInLobe.Add(branch);

                        Vector3 dirToLobe = closestLobe.sphere.Center - branch.end;
                        dirToLobe.Normalize();

                        /*branch.AddChild(new TreeBranch(game,
                            branch, branch.end, 
                            closestLobe.sphere.Center - 
                            dirToLobe * closestLobe.sphere.Radius));*/

                        //allBranches.Add(branch);
                    }
                //}
            }

            foreach (LeafLobe lobe in game.leafCreator.spheres)
            {
                if (lobe.branchesInLobe.Count == 0)
                {
                    TreeBranch parent = branchCreator.treeHierarchy.ParentBranches()[0]; //TODO: get closest
                    //parent.isTrunk = true;
                    //parent.SetBranchType(BranchType.Trunk);
                    Vector3 dirToLobe = lobe.sphere.Center - parent.end;
                    dirToLobe.Normalize();
                    TreeBranch branchToNeglectedLobe = new TreeBranch(game, parent, parent.end, lobe.sphere.Center - 
                            dirToLobe * lobe.sphere.Radius * 1.25f, BranchType.Trunk);
                    branchToNeglectedLobe.controlPoints = lobe.controlPoints;
                    //branchToNeglectedLobe.isTrunk = true;
                    lobe.branchesInLobe.Add(branchToNeglectedLobe);
                    parent.AddChild(branchToNeglectedLobe);
                }
            }

            ////branchCreator.treeHierarchy.setBranches(allBranches);
            game.parentForm.branchViewWindow.RefreshView();
            branchCreator.ResetBranchManager();
        }

        //fill out branches inside leaf lobes using L-System based on input rules
        //partition branches based on number of inputs to each lobe
        //  main branch spawns midway between partitioned terminating control areas
        //TODO: make recursive
        public void FillLeafLobes()
        {
            Console.WriteLine("Filling Leaf Lobes");

            //find areas on exterior of lobe
            game.leafCreator.spheres = game.TreeRules().CreateBranchControlPoints(game.leafCreator.spheres);
            game.leafCreator.spheres = game.TreeRules().CreateLeafControlPoints(game.leafCreator.spheres);

            foreach (LeafLobe lobe in game.leafCreator.spheres)
            {
                for (int i = 0; i < lobe.controlPoints.Count; i++)
                {
                    BoundingSphere controlPoint = lobe.controlPoints[i];
                    float maxY = lobe.sphere.Center.Y + lobe.sphere.Radius;
                    float minY = lobe.sphere.Center.Y - lobe.sphere.Radius;

                    Vector3 dirToCenter = lobe.sphere.Center - controlPoint.Center;
                    dirToCenter.Y = controlPoint.Center.Y;
                    dirToCenter.Normalize();

                    float percentUp = controlPoint.Center.Y / (maxY - minY);

                    //float relaxationFactor = dirToCenter * (percentUp * treeRules.leafTaper);
                    //Vector3 newCenter = new Vector3(relaxationFactor * lobe.sphere.Center.X, controlPoint.Center.Y, relaxationFactor * lobe.sphere.Center.Z);

                    Vector3 TranslationVector = dirToCenter * ((percentUp * treeRules.LeafTaper())) * lobe.sphere.Radius;
                    Vector3 newCenter = controlPoint.Center + TranslationVector;
                    newCenter.Y = controlPoint.Center.Y;
                    //controlPoint.Center = newCenter;

                    lobe.controlPoints[i] = new BoundingSphere(newCenter, controlPoint.Radius);
                }

                //do same taper for leaf control points
                for (int i = 0; i < lobe.leafControlPoints.Count; i++)
                {
                    BoundingSphere controlPoint = lobe.leafControlPoints[i];
                    float maxY = lobe.sphere.Center.Y + lobe.sphere.Radius;
                    float minY = lobe.sphere.Center.Y - lobe.sphere.Radius;

                    Vector3 dirToCenter = lobe.sphere.Center - controlPoint.Center;
                    dirToCenter.Y = controlPoint.Center.Y;
                    dirToCenter.Normalize();

                    float percentUp = controlPoint.Center.Y / (maxY - minY);

                    //float relaxationFactor = dirToCenter * (percentUp * treeRules.leafTaper);
                    //Vector3 newCenter = new Vector3(relaxationFactor * lobe.sphere.Center.X, controlPoint.Center.Y, relaxationFactor * lobe.sphere.Center.Z);

                    Vector3 TranslationVector = dirToCenter * ((percentUp * treeRules.LeafTaper())) * lobe.sphere.Radius;
                    Vector3 newCenter = controlPoint.Center + TranslationVector;
                    newCenter.Y = controlPoint.Center.Y;
                    //controlPoint.Center = newCenter;

                    lobe.leafControlPoints[i] = new BoundingSphere(newCenter, controlPoint.Radius);
                }
            }

            //grow branches
            //partition control points to individual branches
            foreach (LeafLobe lobe in game.leafCreator.spheres)
            {
                foreach (BoundingSphere sphere in lobe.controlPoints)
                {
                    float closestDist = float.MaxValue;
                    TreeBranch closestBranch = null;
                    foreach (TreeBranch branch in lobe.branchesInLobe)
                    {
                        float dist = Vector3.Distance(branch.end, sphere.Center);
                        if (dist < closestDist)
                        {
                            closestDist = dist;
                            closestBranch = branch;
                        }
                    }

                    if (closestBranch != null)
                    {
                        closestBranch.controlPoints.Add(sphere);
                    }
                }

                //do same for leaf control points
                foreach (BoundingSphere sphere in lobe.leafControlPoints)
                {
                    float closestDist = float.MaxValue;
                    TreeBranch closestBranch = null;
                    foreach (TreeBranch branch in lobe.branchesInLobe)
                    {
                        float dist = Vector3.Distance(branch.end, sphere.Center);
                        if (dist < closestDist)
                        {
                            closestDist = dist;
                            closestBranch = branch;
                        }
                    }

                    if (closestBranch != null)
                    {
                        closestBranch.leafControlPoints.Add(sphere);
                    }
                }

                //send branch to midpoint of control points
                foreach (TreeBranch branch in lobe.branchesInLobe)
                {
                    //branch.isTrunk = true;
                    //branch.SetBranchType(BranchType.Trunk);
                    //TODO: reset to just trunk nodes on subsequent clicks
                    //TODO: Num branches based on input rule!
                    //Vector3 end = GetMidpointOfSpheres(branch.controlPoints);
                    //end.Y = lobe.sphere.Center.Y + lobe.sphere.Radius;
                    //TreeBranch connection = new TreeBranch(game, branch, branch.end, end);
                    //connection.controlPoints = branch.controlPoints;
                    //connection.isInsideLobe = branch.isInsideLobe;
                    //branch.AddChild(connection);

                    //CreateBranches(connection);

                    CreateBranches(treeRules.NumTrunkBranchesOffSplit(), branch, treeRules.NumIterations(), treeRules.NumIterations(), BranchType.Trunk);
                }
            }

            game.parentForm.branchViewWindow.RefreshView();
            branchCreator.ResetBranchManager();
        }

        //todo: make recursive
        private void CreateBranches(int numBranches, TreeBranch branch, int iterations, int maxIterations, BranchType branchType)//, Boolean isTrunk = false)
        {
            Console.WriteLine("Create Branches: iteration " + iterations);
            List<BoundingSphere> controlPoints = branch.controlPoints;
            List<BoundingSphere> leafControlPoints = branch.leafControlPoints;
            List<TreeBranch> childBranches = new List<TreeBranch>();

            //if (branch.parent != null)
            //{
                List<BoundingSphere> spheresToRemove = new List<BoundingSphere>();
                foreach (BoundingSphere sphere in branch.controlPoints)
                {
                    if (Vector3.Dot(sphere.Center, branch.start) < 0)
                    {
                        spheresToRemove.Add(sphere);
                    }
                }

                //if (branch.parent != null)
                //{
                //    branch.parent.controlPoints.AddRange(spheresToRemove);
                //}
                /*foreach (BoundingSphere sphere in spheresToRemove)
                {
                    branch.controlPoints.Remove(sphere);
                }*/
            //}

            float degreeRot = 360.0f / (float)(numBranches);

            for (int i = 0; i < numBranches; i++)
            {
                TreeBranch childBranch = new TreeBranch(game, branch, branch.end,
                    Vector3.Distance(branch.leafLobe.sphere.Center + Vector3.Up * branch.leafLobe.sphere.Radius / (float)maxIterations, branch.end), 
                    treeRules.TrunkAngle(), degreeRot * i, branchType);
                childBranch.leafLobe = branch.leafLobe;

                //childBranch.controlPoints = branch.controlPoints;

                //create segmented branch
                //List<TreeBranch> branchSegments = new List<TreeBranch>();
                //for (int j = 0; j < treeRules.numSegmentsInBranch; j++)
                //{
                //    float length = childBranch.Length() / treeRules.numSegmentsInBranch;

                //    TreeBranch segment = null;

                //    if (j == 0)
                //    {
                //        segment = new TreeBranch(game, childBranch.parent, childBranch.start, length, treeRules.segmentAngle, childBranch.isTrunk);
                //    }
                //    else if (j == treeRules.numSegmentsInBranch - 1)
                //    {
                //        segment = new TreeBranch(game, childBranch.parent, branchSegments[branchSegments.Count - 1].end, childBranch.end, childBranch.isTrunk);
                //    }
                //    else
                //    {
                //        segment = new TreeBranch(game, childBranch.parent, branchSegments[branchSegments.Count - 1].end, length, treeRules.segmentAngle, childBranch.isTrunk);
                //    }

                //    if (segment != null)
                //    {
                //        segment.leafLobe = branch.leafLobe;
                //        branchSegments.Add(segment);
                //    }
                //}

                
                childBranches.Add(childBranch);
                //childBranches.AddRange(branchSegments);

                //partition control points!

                //childBranch.controlPoints = branch.controlPoints;
                //childBranch.isInsideLobe = branch.isInsideLobe;
                //branch.AddChild(childBranch);

                //CreateBranches(childBranch);

            }

            foreach (BoundingSphere sphere in controlPoints)
            {
                float closestDist = float.MaxValue;
                TreeBranch closestBranch = null;

                foreach (TreeBranch child in childBranches)
                {
                    float dist = Vector3.Distance(child.end, sphere.Center);
                    if (dist < closestDist)
                    {
                        closestDist = dist;
                        closestBranch = child;
                    }
                }

                if (closestBranch != null)
                {
                    closestBranch.controlPoints.Add(sphere);
                }
            }

            //do same for leaf lobes
            foreach (BoundingSphere sphere in leafControlPoints)
            {
                float closestDist = float.MaxValue;
                TreeBranch closestBranch = null;

                foreach (TreeBranch child in childBranches)
                {
                    float dist = Vector3.Distance(child.end, sphere.Center);
                    if (dist < closestDist)
                    {
                        closestDist = dist;
                        closestBranch = child;
                    }
                }

                if (closestBranch != null)
                {
                    closestBranch.leafControlPoints.Add(sphere);
                }
            }

            //reposition ends
            foreach (TreeBranch child in childBranches)
            {
                //child.isTrunk = true;
                //child.SetBranchType(BranchType.Trunk);
                Vector3 oldEnd = child.end;
                /*child.end = GetMidpointOfSpheres(child.controlPoints);
                if (iterations > 1)
                {
                    child.end = child.end - (child.Dir() * child.Length() / (maxIterations + 1));
                }
                else if (maxIterations == 1)
                {
                    //this is putting the branch too far out
                    child.end += child.Dir() * child.leafLobe.sphere.Radius;
                }*/

                //new attempt
                if (iterations == 1)
                {
                    child.end = GetMidpointOfSpheres(child.controlPoints);

                    if (maxIterations == 1)
                    {
                        //this is putting the branch too far out
                        child.end += child.Dir() * child.leafLobe.sphere.Radius;
                    }
                }

                if (child.end.Y <= child.start.Y)
                {
                    child.end = oldEnd;

                    if (child.leafLobe.sphere.Contains(child.end) == ContainmentType.Disjoint)
                    {
                        float length = Vector3.Distance(child.end, child.leafLobe.sphere.Center);

                        float dif = child.leafLobe.sphere.Radius - length;

                        child.end += child.Dir() * dif;
                    }
                }

                float maxPercent = game.TreeRules().MaxTrunkHeightPercent();
                float rad = child.leafLobe.sphere.Radius;
                float potentialNewY = ((child.leafLobe.sphere.Center.Y - rad) + maxPercent * 2 * rad);
                if (child.end.Y > potentialNewY)
                {
                    child.end.Y = potentialNewY;
                }

                child.isInsideLobe = branch.isInsideLobe;
                child.leafLobe = branch.leafLobe;

                //List<TreeBranch> segments = SegmentBranch(child);

                if (iterations - 1 > 0 && branch.controlPoints.Count > 0)
                {
                    if (branch.parent != null)
                    {
                        branch.parent.controlPoints = new List<BoundingSphere>();
                    }
                    CreateBranches(numBranches, child, iterations - 1, maxIterations, branchType);

                    if (branch.parent != null)
                    {
                        branch.parent.controlPoints.AddRange(spheresToRemove);
                        //CreateBranches(branch.parent);
                    }

                }
                else if (iterations - 1 == 0 || iterations == 0)
                {
                    child.AddLeaves();

                    CreateBranches(child, game.TreeRules().BranchIterations(), child.leafLobe.sphere.Radius / (float)game.TreeRules().BranchIterations(), true);//, treeRules.numIterations);
                }

                //not this one
                //foreach (TreeBranch b in segments)
                //{
                    branch.AddChild(child);
                    //branch.AddChild(b);
                //}
            }

            //Vector3 end = GetMidpointOfSpheres(branch.controlPoints);
            //end.Y = lobe.sphere.Center.Y + lobe.sphere.Radius;
            //TreeBranch connection = new TreeBranch(game, branch, branch.end, end);
            //connection.controlPoints = branch.controlPoints;
            //connection.isInsideLobe = branch.isInsideLobe;
        }

        /// <summary>
        /// take a tree branch from start the end, and segment it
        /// </summary>
        /// <param name="branch">branch from start to end</param>
        /// <returns>parent branch</returns>
        private TreeBranch SegmentBranch(TreeBranch branch)
        {
            List<TreeBranch> branchSegments = new List<TreeBranch>();
            for (int j = 0; j < treeRules.NumSegmentsInBranch(); j++)
            {
                float denominator = treeRules.NumSegmentsInBranch() - 1;
                if (denominator == 0)
                    denominator = 1;
                float length = branch.Length() / denominator;

                TreeBranch segment = null;

                if (j == 0)
                {
                    Vector3 end = branch.start + branch.Dir() * length;
                    end += new Vector3((float)(game.rand.NextDouble() * 2 - 1) * (length / 2),
                        (float)(game.rand.NextDouble() * 2 - 1) * (length / 2),
                        (float)(game.rand.NextDouble() * 2 - 1) * (length / 2));
                    //end.Y += (float)((game.rand.NextDouble() * 2) - 1) * treeRules.segmentOffest * length;
                    //segment = new TreeBranch(game, branch.parent, branch.start, end, branch.isTrunk);

                    segment = new TreeBranch(game, branch/*.parent*/, branch.start, end, BranchType.Branch);
                }
                else if (j == treeRules.NumSegmentsInBranch() - 1)
                {
                    segment = new TreeBranch(game, branch.parent, branchSegments[branchSegments.Count - 1].end, branch.end, branch.branchType);
                }
                else
                {
                    Vector3 end = branchSegments[branchSegments.Count - 1].end + branch.Dir() * length;
                    end += new Vector3((float)(game.rand.NextDouble() * 2 - 1) * (length / 2),
                        (float)(game.rand.NextDouble() * 2 - 1) * (length / 2),
                        (float)(game.rand.NextDouble() * 2 - 1) * (length / 2));
                    //end.Y += (float)((game.rand.NextDouble() * 2) - 1) * treeRules.segmentOffest * length;
                    segment = new TreeBranch(game, branchSegments[branchSegments.Count - 1], branchSegments[branchSegments.Count - 1].end, end, BranchType.Branch);

                    branchSegments[branchSegments.Count - 1].AddChild(segment);
                }

                if (segment != null)
                {
                    //segment.isTrunk = false;
                    //segment.SetBranchType(BranchType.Branch);
                    segment.leafLobe = branch.leafLobe;
                    segment.AddCap();
                    branchSegments.Add(segment);
                }
            }

            return branchSegments[0];
        }

        List<Vector3> starts = new List<Vector3>();
        private void CreateBranches(TreeBranch parent, int iteration, float length, Boolean firstPass)
        {
            TreeBranch newBranch = null;
            if (firstPass)
            {
                foreach (BoundingSphere sphere in parent.controlPoints)
                {
                    newBranch = null;
                    float angleToControlPoint = game.TreeRules().Angle();

                    Vector3 A = parent.start - sphere.Center;
                    A.Normalize();
                    Vector3 B = parent.start - parent.end;
                    B.Normalize();
                    float startToEndAngle = (float)Math.Acos(Vector3.Dot(A, B));
                    float endtoEndAngle = (float)Math.Acos((double)(Vector3.Dot(sphere.Center - parent.end, parent.end - parent.start)));

                    //if (parent.isInsideLobe)
                    //{
                        newBranch = InterpolateTreeBranchOnParent(length, sphere, parent, startToEndAngle, angleToControlPoint);
                    //}
                    /*else if (startToEndAngle >= angleToControlPoint) //branch too far beyond start
                    {
                        newBranch = new TreeBranch(game, parent, parent.start, sphere.Center, BranchType.Branch);
                    }
                    else if (endtoEndAngle <= angleToControlPoint) //branch too far beyond end
                    {
                        //newBranch = new TreeBranch(game, parent, parent.end, sphere.Center);

                        newBranch = InterpolateTreeBranchOnParent(sphere, parent, startToEndAngle, angleToControlPoint);
                    }
                    else //branch within parent bound
                    {
                        newBranch = InterpolateTreeBranchOnParent(sphere, parent, startToEndAngle, angleToControlPoint);
                    }*/


                    if (newBranch != null)
                    {

                        if (newBranch.end.Y >= newBranch.start.Y)//parent.start.Y)//newBranch.start.Y)
                        {
                            //List<TreeBranch> segments = SegmentBranch(newBranch);

                            //foreach (TreeBranch segment in segments)
                            //{

                            if (!starts.Contains(newBranch.start))
                            {
                                starts.Add(newBranch.start);

                                TreeBranch segmentedBranch = SegmentBranch(newBranch);

                                if (segmentedBranch.Dir() != parent.Dir())
                                {
                                    parent.AddChild(segmentedBranch);

                                    if (iteration - 1 > 0)
                                    {
                                        CreateBranches(GetFirstTerminatingBranch(segmentedBranch), iteration - 1, length * game.TreeRules().LengthReductionFactor(), false);
                                    }
                                }
                            }

                            else
                            {
                                List<TreeBranch> c = parent.children;
                                List<TreeBranch> copiedChildren = new List<TreeBranch>();
                                foreach (TreeBranch branch in c)
                                {
                                    copiedChildren.Add(branch);
                                }
                                foreach (TreeBranch branch in copiedChildren)
                                {
                                    if (branch.start == newBranch.start)
                                    {
                                        if (Math.Acos(Vector3.Dot(newBranch.Dir(), branch.Dir())) > 2)
                                        {
                                            starts.Add(newBranch.start);

                                            TreeBranch segmentedBranch = SegmentBranch(newBranch);

                                            if (segmentedBranch.Dir() != parent.Dir())
                                            {
                                                parent.AddChild(segmentedBranch);

                                                if (iteration - 1 > 0)
                                                {
                                                    CreateBranches(GetFirstTerminatingBranch(segmentedBranch), iteration - 1, length * game.TreeRules().LengthReductionFactor(), false);
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            //}

                            //parent.AddChild(newBranch);

                            //if (iteration - 1 > 0)
                            //{
                            //    CreateBranches(newBranch, iteration - 1);
                            //}
                        }
                    }
                }
            }
            else
            {
                int numBranchesOffSplit = game.TreeRules().NumBranchesOffSplit();
                for (int i = 0; i < numBranchesOffSplit; i++)
                {
                    float rotateAround = i * (360.0f / (float)numBranchesOffSplit);
                    newBranch = new TreeBranch(game, parent, parent.end, length, game.TreeRules().SubsequentAngle(), rotateAround, BranchType.Branch);

                    TreeBranch segmentedBranch = SegmentBranch(newBranch);

                //TreeBranch lastBranch;
                //TreeBranch segmentedBranch = SegmentBranch(newBranch, out lastBranch);

                //if (segmentedBranch.Dir() != parent.Dir())
                //{
                    parent.AddChild(segmentedBranch);

                    if (iteration - 1 > 0)
                    {
                        CreateBranches(GetFirstTerminatingBranch(segmentedBranch), iteration - 1, length * game.TreeRules().LengthReductionFactor(), false);
                    }
                //}
                }
            }

            game.parentForm.branchViewWindow.RefreshView();
            branchCreator.ResetBranchManager();
        }

        private TreeBranch GetFirstTerminatingBranch(TreeBranch parent)
        {
            if (parent.children.Count > 0)
            {
                if (parent.children[0].children.Count == 0)
                {
                    return parent.children[0];
                }
                else
                {
                    return GetFirstTerminatingBranch(parent.children[0]);
                }
            }

            return null;
        }

        private TreeBranch InterpolateTreeBranchOnParent(float length, BoundingSphere sphere, TreeBranch parent, float startToEndAngle, float angleToControlPoint)
        {
            float lengthFromStartToControl = Vector3.Distance(sphere.Center, parent.start);
            float h = lengthFromStartToControl * (float)Math.Sin(startToEndAngle);

            float epsilon = lengthFromStartToControl * (float)Math.Cos(startToEndAngle) - parent.Length();

            float lengthToControl = h / (float)Math.Sin(angleToControlPoint);
            float translationDistanceAlongParent = lengthToControl * (float)Math.Cos(angleToControlPoint) - epsilon;

            Vector3 start = parent.end - parent.Dir() * translationDistanceAlongParent;

            if (translationDistanceAlongParent > parent.Length())
            {
                start = parent.start;
            }
            else if (translationDistanceAlongParent < 0)
            {
                start = parent.end;
            }

            length = MathHelper.Clamp(length, 0, Vector3.Distance(start, sphere.Center));

            Vector3 dir = sphere.Center - start;
            dir.Normalize();

            Vector3 end = start + dir * length;

            return new TreeBranch(game, parent, start, end, BranchType.Branch);
        }

        private Vector3 GetMidpointOfSpheres(List<BoundingSphere> spheres)
        {
            Vector3 mid = Vector3.Zero;

            foreach (BoundingSphere sphere in spheres)
            {
                mid += sphere.Center;
            }

            mid /= spheres.Count;

            return mid;
        }

        //create leaves from terminating branches to edge of leaf lobes
        public void CreateLeaves()
        {
            game.TreeRules().ConnectBranchesToLeafLobes();

            //List<TreeBranch> branches = game.branchCreator.treeHierarchy.GetTerminatingBranches();

            //foreach (LeafLobe lobe in game.leafCreator.spheres)
            //{
            //    foreach (BoundingSphere leafControlPoint in lobe.leafControlPoints)
            //    {


            //        foreach (TreeBranch branch in branches)
            //        {

            //        }
            //    }
            //}

/*            Console.WriteLine("Creating Leaves");

            //First create final branches for leaves
            //for (int i = 0; i < treeRules.numIterations; i++)
            //{
                game.parentForm.branchViewWindow.RefreshView();
                branchCreator.ResetBranchManager();

                List<TreeBranch> terminatingBranches = branchCreator.treeHierarchy.GetTerminatingBranches();

                for (int j = 0; j < treeRules.NumFinalBranches(); j++)
                {
                    foreach (TreeBranch branch in terminatingBranches)
                    {
                        if (branch.isTrunk)
                        {
                            branch.isTrunk = false;
                            branch.AddLeaves();
                        }
                        CreateLeafBranches(branch, treeRules.subsequentAngle);
                    }
                }
            //}

            game.parentForm.branchViewWindow.RefreshView();
            branchCreator.ResetBranchManager();
*/
            //then add leaves to branches
        }

        private void CreateLeafBranches(TreeBranch parent, float angle)
        {
            TreeBranch newBranch = new TreeBranch(game, parent, parent.GetPositionOnBranch(), parent.Length() * treeRules.LengthReductionFactor(), angle, BranchType.Branch);
            parent.AddChild(newBranch);
        }

        public void CreateGeometry()
        {
            foreach (TreeBranch branch in branchCreator.treeHierarchy.GetAllBranches())
            {
                //branch.Initialize(branch.parent, branch.start, branch.end, branch.id, false);

                if (branch.addLeaves)
                {
                    branch.AddLeaves();
                }
                branch.CreateVolume();
            }

            InitializePrimitive(game.GraphicsDevice);
        }

        public void Update()
        {
            branchCreator.Update();
        }

        public void AddLeaf(Quad leaf, Texture2D texture)
        {
            leafQuads.Add(new LeafTexture(leaf, texture));
        }

        public void Draw()
        {
            if (game.showBoundingVolumes)
            {
                branchCreator.Draw();

                foreach (LeafLobe lobe in game.leafCreator.spheres)
                {
                    foreach (BoundingSphere point in lobe.controlPoints)
                    {
                        BoundingSphereRenderer.Render(point, game.GraphicsDevice, game.camera.View, game.camera.Projection, Color.Purple);
                    }

                    foreach (BoundingSphere point in lobe.leafControlPoints)
                    {
                        BoundingSphereRenderer.Render(point, game.GraphicsDevice, game.camera.View, game.camera.Projection, Color.Green);
                    }
                }
            }

            foreach (TreeBranch branch in branchCreator.treeHierarchy.ParentBranches())
            {
                branch.Draw();
            }

            //
            DrawTrunk();
            if (game.drawLeaves)
            {
                #region leaves
                game.GraphicsDevice.BlendState = BlendState.AlphaBlend;
                //Draw Leaves
                foreach (LeafTexture leafTex in leafQuads)
                {
                    foreach (EffectPass pass in quadEffect.CurrentTechnique.Passes)
                    {
                        pass.Apply();

                        quadEffect.World = leafTex.leaf.World;//Matrix.Identity;
                        quadEffect.View = game.camera.View;
                        quadEffect.Projection = game.camera.Projection;
                        quadEffect.TextureEnabled = true;
                        quadEffect.Texture = leafTex.tex;

                        game.GraphicsDevice.DrawUserIndexedPrimitives
                            <VertexPositionNormalTexture>(
                            PrimitiveType.TriangleList,
                            leafTex.leaf.Vertices, 0, 4,
                            leafTex.leaf.Indexes, 0, 2);
                    }
                }
                game.GraphicsDevice.BlendState = BlendState.Opaque;
                #endregion
                DrawTrunk();
            }
        }

        /// <summary>
        /// Once all the geometry has been specified by calling AddVertex and AddIndex,
        /// this method copies the vertex and index data into GPU format buffers, ready
        /// for efficient rendering.
        public void InitializePrimitive(GraphicsDevice graphicsDevice)
        {
            try
            {
                // Create a vertex declaration, describing the format of our vertex data.

                // Create a vertex buffer, and copy our vertex data into it.
                branchVertexBuffer = new VertexBuffer(graphicsDevice,
                                                typeof(VertexPositionNormal),
                                                branchVertices.Count, BufferUsage.None);

                branchVertexBuffer.SetData(branchVertices.ToArray());

                // Create an index buffer, and copy our index data into it.
                branchIndexBuffer = new IndexBuffer(graphicsDevice, typeof(int),
                                              branchIndices.Count, BufferUsage.None);

                branchIndexBuffer.SetData(branchIndices.ToArray());

                // Create a BasicEffect, which will be used to render the primitive.
                branchBasicEffect = new BasicEffect(graphicsDevice);

                branchBasicEffect.EnableDefaultLighting();
            }
            catch { }
        }

        private void DrawTrunk()
        {
            #region trunk
            if (branchBasicEffect != null)
            {
                // Set BasicEffect parameters.
                branchBasicEffect.World = Matrix.Identity;
                branchBasicEffect.View = game.camera.View;
                branchBasicEffect.Projection = game.camera.Projection;
                branchBasicEffect.DiffuseColor = Color.SaddleBrown.ToVector3();
                branchBasicEffect.Alpha = Color.SaddleBrown.A / 255.0f;
                branchBasicEffect.TextureEnabled = true;
                branchBasicEffect.Texture = game.bark;

                GraphicsDevice device = branchBasicEffect.GraphicsDevice;
                device.DepthStencilState = DepthStencilState.Default;

                if (Color.SaddleBrown.A < 255)
                {
                    // Set renderstates for alpha blended rendering.
                    device.BlendState = BlendState.AlphaBlend;
                }
                else
                {
                    // Set renderstates for opaque rendering.
                    device.BlendState = BlendState.Opaque;
                }

                // Set our vertex declaration, vertex buffer, and index buffer.
                branchBasicEffect.GraphicsDevice.SetVertexBuffer(branchVertexBuffer);

                branchBasicEffect.GraphicsDevice.Indices = branchIndexBuffer;


                foreach (EffectPass effectPass in branchBasicEffect.CurrentTechnique.Passes)
                {
                    effectPass.Apply();

                    int primitiveCount = branchIndices.Count / 3;

                    branchBasicEffect.GraphicsDevice.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0,
                                                         branchVertices.Count, 0, primitiveCount);

                }
            }
            #endregion
        }

    }
}
